@@ -3,7 +3,7 @@ package Net::DNS::Nameserver;
 use strict;
 use warnings;
 
-our $VERSION = (qw$Id: Nameserver.pm 1925 2023-05-31 11:58:59Z willem $)[2];
+our $VERSION = (qw$Id: Nameserver.pm 2002 2025-01-07 09:57:46Z willem $)[2];
 
 
 =head1 NAME
@@ -12,18 +12,22 @@ Net::DNS::Nameserver - DNS server class
 
 =head1 SYNOPSIS
 
-    use Net::DNS::Nameserver;
+	use Net::DNS::Nameserver;
 
-    my $nameserver = Net::DNS::Nameserver->new(
-	LocalAddr	=> ['::1', '127.0.0.1'],
-	ZoneFile	=> "filename"
-	);
+	my $nameserver = Net::DNS::Nameserver->new(
+			LocalAddr	=> ['::1', '127.0.0.1'],
+			LocalPort	=> 15353,
+			ZoneFile	=> 'filename'
+			);
 
-    my $nameserver = Net::DNS::Nameserver->new(
-	LocalAddr	=> '10.1.2.3',
-	LocalPort	=> 5353,
-	ReplyHandler	=> \&reply_handler
-    );
+	my $nameserver = Net::DNS::Nameserver->new(
+			LocalAddr	=> '10.1.2.3',
+			LocalPort	=> 15353,
+			ReplyHandler	=> \&reply_handler
+			);
+
+	$nameserver->start_server($timeout);
+	$nameserver->stop_server;
 
 
 =head1 DESCRIPTION
@@ -35,7 +39,7 @@ queries emanating from a client resolver.
 It is not, nor will it ever be, a general-purpose DNS nameserver
 implementation.
 
-See L</EXAMPLE> for an example.
+See L</EXAMPLES> below for further details.
 
 =cut
 
@@ -44,15 +48,27 @@ use Carp;
 use Net::DNS;
 use Net::DNS::ZoneFile;
 
-use IO::Socket::IP 0.38;
 use IO::Select;
+use IO::Socket::IP;
+use IO::Socket;
+use Socket;
 
-use constant USE_POSIX => defined eval 'use POSIX ":sys_wait_h"; 1';	## no critic
+use constant SOCKOPT => eval {
+	my @sockopt;
+	push @sockopt, eval '[SOL_SOCKET, SO_REUSEADDR]';	## no critic
+	push @sockopt, eval '[SOL_SOCKET, SO_REUSEPORT]';	## no critic
+
+	my $filter = sub {					# check that options safe to use
+		return eval { IO::Socket::IP->new( Proto => "udp", Sockopts => [shift], Type => SOCK_DGRAM ) }
+	};
+	return grep { &$filter($_) } @sockopt;			# without any guarantee that they work!
+};
 
 use constant DEFAULT_ADDR => qw(::1 127.0.0.1);
-use constant DEFAULT_PORT => 5353;
+use constant DEFAULT_PORT => 15353;
 
-use constant PACKETSZ => 512;
+use constant POSIX => defined eval 'use POSIX ":sys_wait_h"; 1';	## no critic
+use constant MSWin => scalar( $^O =~ /MSWin/i );
 
 
 #------------------------------------------------------------------------------
@@ -60,33 +76,35 @@ use constant PACKETSZ => 512;
 #------------------------------------------------------------------------------
 
 sub new {
-	my ( $class, %self ) = @_;
+	my ( $class, %config ) = @_;
+
+	my %self = (
+		LocalAddr => [DEFAULT_ADDR],
+		LocalPort => [DEFAULT_PORT],
+		Truncate  => 1,
+		%config
+		);
 	my $self = bless \%self, $class;
 
-	$self->ReadZoneFile( $self{ZoneFile} ) if exists $self{ZoneFile};
+	$self->_ReadZoneFile( $self{ZoneFile} ) if exists $self{ZoneFile};
 
 	croak 'No reply handler!' unless ref( $self{ReplyHandler} ) eq "CODE";
 
 	# local server addresses need to be accepted by a resolver
-	my $LocalAddr = $self{LocalAddr} || [DEFAULT_ADDR];
+	my $LocalAddr = $self{LocalAddr};
 	my $resolver  = Net::DNS::Resolver->new( nameservers => $LocalAddr );
-	$resolver->force_v4(1) if $self{Force_IPv4};
-	$resolver->force_v6(1) if $self{Force_IPv6};
+	$resolver->force_v4( $self{Force_IPv4} );
+	$resolver->force_v6( $self{Force_IPv6} );
 	$self{LocalAddr} = [$resolver->nameservers];
-	$self{LocalPort} ||= DEFAULT_PORT;
-
-	$self{Truncate}	   = 1	 unless defined( $self{Truncate} );
-	$self{IdleTimeout} = 120 unless defined( $self{IdleTimeout} );
-
 	return $self;
 }
 
 
 #------------------------------------------------------------------------------
-# ReadZoneFile - Read zone file used by default reply handler
+# _ReadZoneFile - Read zone file used by default reply handler
 #------------------------------------------------------------------------------
 
-sub ReadZoneFile {
+sub _ReadZoneFile {
 	my ( $self, $file ) = @_;
 	my $zonefile = Net::DNS::ZoneFile->new($file);
 
@@ -107,16 +125,16 @@ sub ReadZoneFile {
 
 	$self->{namelist}     = [sort { length($b) <=> length($a) } keys %$RRhash];
 	$self->{zonelist}     = [sort { length($b) <=> length($a) } @zonelist];
-	$self->{ReplyHandler} = sub { $self->ReplyHandler(@_) };
+	$self->{ReplyHandler} = sub { $self->_ReplyHandler(@_) };
 	return;
 }
 
 
 #------------------------------------------------------------------------------
-# ReplyHandler - Default reply handler serving RRs from zone file
+# _ReplyHandler - Default reply handler serving RRs from zone file
 #------------------------------------------------------------------------------
 
-sub ReplyHandler {
+sub _ReplyHandler {
 	my ( $self, $qname, $qclass, $qtype, $peerhost, $query, $conn ) = @_;
 	my $RRhash = $self->{index};
 	my $rcode;
@@ -151,9 +169,8 @@ sub ReplyHandler {
 				s/^\*//;			# delete leading asterisk
 				s/([.?*+])/\\$1/g;		# escape dots and regex quantifiers
 				next unless $qname =~ /[.]?([^.]+$_)$/i;
-				my $encloser = $1;		# check no ENT encloses qname
-				$rcode = 'NOERROR';
-				last if grep {/(^|\.)$encloser$/i} @RRname;    # [NODATA]
+				my $cover = $1;			# check for name covering wildcard
+				next if grep {/[.]?$cover$/i} @RRname;
 
 				my ($q) = $query->question;	# synthesise RR at qname
 				foreach my $rr ( @{$RRhash->{$wildcard}} ) {
@@ -161,6 +178,7 @@ sub ReplyHandler {
 					$clone->{owner} = $q->{qname};
 					push @match, $clone;
 				}
+				$rcode = 'NOERROR';
 				last;
 			}
 		}
@@ -171,9 +189,10 @@ sub ReplyHandler {
 		push @ans, grep { $_->type eq $qtype } @match;
 		unless (@ans) {
 			foreach ( @{$self->{zonelist}} ) {
+				my $RRlist = $RRhash->{lc $_};
 				s/([.?*+])/\\$1/g;		# escape dots and regex quantifiers
 				next unless $qname =~ /[^.]+[.]$_[.]?$/i;
-				push @auth, grep { $_->type eq 'SOA' } @{$RRhash->{$_}};
+				push @auth, grep { $_->type eq 'SOA' } @$RRlist;
 				last;
 			}
 		}
@@ -184,10 +203,10 @@ sub ReplyHandler {
 
 
 #------------------------------------------------------------------------------
-# make_reply - Make a reply packet.
+# _make_reply - Make a reply packet.
 #------------------------------------------------------------------------------
 
-sub make_reply {
+sub _make_reply {
 	my ( $self, $query, $sock ) = @_;
 	my $verbose = $self->{Verbose};
 
@@ -288,95 +307,86 @@ sub make_reply {
 
 
 #------------------------------------------------------------------------------
-# TCP_connection - Handle a TCP connection.
+# _TCP_connection - Handle a TCP connection.
 #------------------------------------------------------------------------------
 
-sub TCP_connection {
-	my ( $self, $socket ) = @_;
-	my $timeout = $self->{IdleTimeout};
+sub _TCP_connection {
+	my ( $self, $socket, $buffer ) = @_;
 	my $verbose = $self->{Verbose};
 
-	while (1) {
-		alarm $timeout;
-		my $l = '';
-		my $n = sysread( $socket, $l, 2 );
-		unless ( defined $n ) {
-			redo if $!{EINTR};	## retry if aborted by signal
-			die "sysread: $!";
-		}
-		last if $n == 0;
-		my $msglen = unpack 'n', $l;
-
-		my $buffer = '';
-		while ( $msglen > ( my $l = length $buffer ) ) {
-			my $n = sysread( $socket, $buffer, ( $msglen - $l ), $l );
-			unless ( defined $n ) {
-				redo if $!{EINTR};	## retry if aborted by signal
-				die "sysread: $!";
-			}
-		}
-
-		if ($verbose) {
-			my $peer = $socket->peerhost;
-			my $port = $socket->peerport;
-			my $size = length $buffer;
-			print "Received $size octets from [$peer] port $port\n";
-		}
-
-		my $query = Net::DNS::Packet->new( \$buffer );
-		if ($@) {
-			print "Error decoding query packet: $@\n" if $verbose;
-			undef $query;	## force FORMERR reply
-		}
-
-		my $reply = $self->make_reply( $query, $socket );
-		die 'Failed to create reply' unless defined $reply;
-
-		my $segment = $reply->data(65500);		# limit to one TCP envelope
-		my $length  = length $segment;
-		warn "Multi-packet TCP response not implemented" if $reply->header->tc;
-		if ($verbose) {
-			print "TCP response (2 + $length octets) - ";
-			print $socket->send( pack 'na*', $length, $segment ) ? "sent" : "failed: $!", "\n";
-		} else {
-			$socket->send( pack 'na*', $length, $segment );
-		}
-	}
-	alarm 0;
-	close $socket;
-	return;
-}
-
-
-#------------------------------------------------------------------------------
-# UDP_connection - Handle a UDP connection.
-#------------------------------------------------------------------------------
-
-sub UDP_connection {
-	my ( $self, $socket ) = @_;
-	my $verbose = $self->{Verbose};
-	alarm 3;
-
-	my $buffer = "";
-	return unless defined $socket->recv( $buffer, PACKETSZ );
-
-	if ($verbose) {
-		my $peer = $socket->peerhost;
-		my $port = $socket->peerport;
-		my $size = length $buffer;
-		print "Received $size octets from [$peer] port $port\n";
-	}
-
 	my $query = Net::DNS::Packet->new( \$buffer );
 	if ($@) {
 		print "Error decoding query packet: $@\n" if $verbose;
 		undef $query;		## force FORMERR reply
 	}
 
-	my $reply = $self->make_reply( $query, $socket );
+	my $reply = $self->_make_reply( $query, $socket );
 	die 'Failed to create reply' unless defined $reply;
 
-	my @UDPsize = ( $query && $self->{Truncate} ) ? $query->edns->UDPsize : ();
+	my $segment = $reply->data;
+	my $length  = length $segment;
+	if ($verbose) {
+		print "TCP response (2 + $length octets) - ";
+		print $socket->send( pack 'na*', $length, $segment ) ? "sent" : "failed: $!", "\n";
+	} else {
+		$socket->send( pack 'na*', $length, $segment );
+	}
+	return;
+}
+
+sub _read_tcp {
+	my ( $socket, $verbose ) = @_;
+
+	my $header = '';
+	local $! = 0;
+	my $n = sysread( $socket, $header, 2 );
+	unless ( defined $n ) {
+		redo if $!{EINTR};	## retry if aborted by signal
+		die "sysread: $!";
+	}
+	return '' if $n == 0;
+	return '' if length($header) < 2;
+	my $msglen = unpack 'n', $header;
+
+	my $buffer = '';
+	while ( $msglen > ( my $len = length $buffer ) ) {
+		local $! = 0;
+		my $n = sysread( $socket, $buffer, ( $msglen - $len ), $len );
+		unless ( defined $n ) {
+			redo if $!{EINTR};	## retry if aborted by signal
+			die "sysread: $!";
+		}
+		last if $n == 0;	## client closed (or lied)	per RT#151240
+	}
+
+	if ($verbose) {
+		my $peer = $socket->peerhost;
+		my $port = $socket->peerport;
+		my $size = length $buffer;
+		print "Received $size octets from [$peer] port $port\n";
+	}
+	return $buffer;
+}
+
+
+#------------------------------------------------------------------------------
+# _UDP_connection - Handle a UDP connection.
+#------------------------------------------------------------------------------
+
+sub _UDP_connection {
+	my ( $self, $socket, $buffer ) = @_;
+	my $verbose = $self->{Verbose};
+
+	my $query = Net::DNS::Packet->new( \$buffer );
+	if ($@) {
+		print "Error decoding query packet: $@\n" if $verbose;
+		undef $query;		## force FORMERR reply
+	}
+
+	my $reply = $self->_make_reply( $query, $socket );
+	die 'Failed to create reply' unless defined $reply;
+
+	my @UDPsize = ( $query && $self->{Truncate} ) ? $query->edns->UDPsize || 512 : ();
 	if ($verbose) {
 		my $response = $reply->data(@UDPsize);
 		print 'UDP response (', length($response), ' octets) - ';
@@ -384,189 +394,185 @@ sub UDP_connection {
 	} else {
 		$socket->send( $reply->data(@UDPsize) );
 	}
-	alarm 0;
-	close $socket;
 	return;
 }
 
+sub _read_udp {
+	my ( $socket, $verbose ) = @_;
+	my $buffer = '';
+	$socket->recv( $buffer, 9000 );	## payload limit for Ethernet "Jumbo" packet
+	if ($verbose) {
+		my $peer = $socket->peerhost;
+		my $port = $socket->peerport;
+		my $size = length $buffer;
+		print "Received $size octets from [$peer] port $port\n";
+	}
+	return $buffer;
+}
+
 
 #------------------------------------------------------------------------------
 # Socket mechanics.
 #------------------------------------------------------------------------------
 
-use constant DEBUG => 0;
+use constant DEBUG => $ENV{DEBUG} ? 1 : 0;
 
-sub logmsg { return print "$0 $$: @_ at ", scalar localtime(), "\n" }
+sub _logmsg { warn( join '', "$0 $$: @_ at ", scalar localtime(), "\n" ); return }
 
-sub spawn {
-	my $coderef = shift;
-	confess "usage: spawn CODEREF" unless ref($coderef) eq 'CODE';
-
-	unless ( defined( my $pid = fork() ) ) {
-		die "cannot fork: $!";
-	} elsif ($pid) {
-		logmsg "begat $pid" if DEBUG;
-		return $pid;		## parent
-	}
-
-	# else ...			## child
-	$coderef->();
-	exit;
-}
-
-sub reaper {
-	local $!;			## protect current error
-	$SIG{CHLD} = \&reaper;		## no critic	sysV semantics
-	while ( ( my $pid = waitpid( -1, USE_POSIX ? WNOHANG : 0 ) ) > 0 ) {
-		logmsg "reaped $pid" if DEBUG;
-	}
-	return;
-}
-
-
-sub start_server {
-	my $self = shift;
-	my $list = $self->{LocalAddr};
-	my $port = $self->{LocalPort};
-	my @pid;
-	foreach my $ip (@$list) {
-		push @pid, spawn( sub { $self->TCP_server( $ip, $port ) } );
-		push @pid, spawn( sub { $self->UDP_server( $ip, $port ) } );
-	}
-	return @pid;
-}
-
-sub start_noloop {
-	my ( $self, $timeout ) = ( @_, 600 );
-	my $list = $self->{LocalAddr};
-	my $port = $self->{LocalPort};
-	foreach my $ip (@$list) {
-		spawn(	sub {
-				alarm $timeout;
-				$self->TCP_initialise( $ip, $port );
-			} );
-		spawn(	sub {
-				alarm $timeout;
-				$self->UDP_initialise( $ip, $port );
-			} );
-	}
-	return;
-}
-
-
-sub TCP_initialise {
-	my ( $self, $ip, $port ) = @_;
+sub _TCP_socket {
+	my ( $ip, $port ) = @_;
 	my $socket = IO::Socket::IP->new(
 		LocalAddr => $ip,
 		LocalPort => $port,
-		ReuseAddr => 1,
-		ReusePort => 1,
+		Sockopt	  => [SOCKOPT],
 		Proto	  => "tcp",
 		Listen	  => SOMAXCONN,
 		Type	  => SOCK_STREAM
 		)
-			|| die "can't setup TCP socket";
+			or die "can't setup TCP socket: $!";
 
-	logmsg "TCP server [$ip] port $port started";
-
-	{
-		my $client = $socket->accept() || do {
-			redo if $!{EINTR};	## retry if aborted by signal
-			die "accept: $!";
-		};
-
-		spawn( sub { $self->TCP_connection($client) } );
-	}
+	_logmsg "TCP server [$ip] port $port" if DEBUG;
 	return $socket;
 }
 
-sub TCP_server {
-	my ( $self, $ip, $port ) = @_;
-	my $socket = $self->TCP_initialise( $ip, $port );
-	while (1) {
-		my $client = $socket->accept() || do {
+sub _TCP_server {
+	my ( $self, $ip, $port, $timeout ) = @_;
+	my $listen = _TCP_socket( $ip, $port );
+	my $select = IO::Select->new($listen);
+
+	my $expired;
+	my $terminate = sub { $expired++ };
+	local $SIG{ALRM} = $terminate;
+	local $SIG{TERM} = $terminate;
+	alarm $timeout;
+	until ($expired) {
+		local $! = 0;
+		scalar( my @ready = $select->can_read(2) ) or do {
 			redo if $!{EINTR};	## retry if aborted by signal
-			die "accept: $!";
+			last if $!;
 		};
 
-		spawn( sub { $self->TCP_connection($client) } );
+		foreach my $socket (@ready) {
+			if ( $socket == $listen ) {
+				$select->add( $listen->accept );
+				next;
+			}
+			if ( my $buffer = _read_tcp( $socket, $self->{Verbose} ) ) {
+				_spawn( sub { $self->_TCP_connection( $socket, $buffer ) } );
+			} else {
+				close($socket);
+				$select->remove($socket);
+			}
+		}
+		sleep(0) if MSWin;
 	}
-	exit;
+	return;
 }
 
 
-sub UDP_initialise {
-	my ( $self, $ip, $port ) = @_;
+sub _UDP_socket {
+	my ( $ip, $port ) = @_;
 	my $socket = IO::Socket::IP->new(
 		LocalAddr => $ip,
 		LocalPort => $port,
-		ReuseAddr => 1,
-		ReusePort => 1,
+		Sockopt	  => [SOCKOPT],
 		Proto	  => "udp",
 		Type	  => SOCK_DGRAM
 		)
-			|| die "can't setup UDP socket";
+			or die "can't setup UDP socket: $!";
 
-	logmsg "UDP server [$ip] port $port started";
-
-	my $select = IO::Select->new($socket);
-	{
-		local $! = 0;
-		scalar( my @ready = $select->can_read() ) || do {
-			redo if $!{EINTR};	## retry if aborted by signal
-			die "select->can_read(): $!";
-		};
-
-		foreach my $client (@ready) {
-			spawn( sub { $self->UDP_connection($client) } );
-		}
-	}
+	_logmsg "UDP server [$ip] port $port" if DEBUG;
 	return $socket;
 }
 
-sub UDP_server {
-	my ( $self, $ip, $port ) = @_;
-	my $socket = $self->UDP_initialise( $ip, $port );
+sub _UDP_server {
+	my ( $self, $ip, $port, $timeout ) = @_;
+	my $socket = _UDP_socket( $ip, $port );
 	my $select = IO::Select->new($socket);
-	while (1) {
+
+	my $expired;
+	my $terminate = sub { $expired++ };
+	local $SIG{ALRM} = $terminate;
+	local $SIG{TERM} = $terminate;
+	alarm $timeout;
+	until ($expired) {
 		local $! = 0;
-		scalar( my @ready = $select->can_read() ) || do {
+		scalar( my @ready = $select->can_read(2) ) or do {
 			redo if $!{EINTR};	## retry if aborted by signal
-			die "select->can_read(): $!";
+			last if $!;
 		};
 
 		foreach my $client (@ready) {
-			spawn( sub { $self->UDP_connection($client) } );
+			my $buffer = _read_udp( $client, $self->{Verbose} );
+			_spawn( sub { $self->_UDP_connection( $client, $buffer ) } );
 		}
+		sleep(0) if MSWin;
 	}
-	exit;
-}
-
-
-#------------------------------------------------------------------------------
-# main_loop - Start nameserver loop.
-#------------------------------------------------------------------------------
-
-sub main_loop {
-	local $SIG{CHLD} = \&reaper;
-	my @pid = shift->start_server;
-	local $SIG{TERM} = sub { kill( 'TERM', @pid ) };
-	1 while waitpid( -1, 0 ) > 0;
-	logmsg "@pid terminated";
 	return;
 }
 
 
 #------------------------------------------------------------------------------
-# loop_once - Start single-transaction nameserver
+# Process mechanics.
 #------------------------------------------------------------------------------
 
-sub loop_once {
-	my ( $self, @timeout ) = @_;
-	local $SIG{CHLD} = \&reaper;
-	$self->start_noloop(@timeout);
-	1 while waitpid( -1, 0 ) > 0;	## park main process until timeout or
-	return;				## user CTRL_C kills remaining children
+my $noop = sub { };
+
+sub _spawn {
+	my $coderef = shift;
+	unless ( defined( my $pid = fork() ) ) {
+		die "cannot fork: $!";
+	} elsif ($pid) {
+		_logmsg "begat $pid" if DEBUG;
+		return $pid;		## parent
+	}
+
+	# else ...
+	local $SIG{TERM} = $noop;
+	local $SIG{CHLD} = \&_reaper;
+	$coderef->();			## child
+	exit;
+}
+
+sub _reaper {
+	local ( $!, $? );		## protect error and exit status
+	$SIG{CHLD} = \&_reaper;		## no critic	sysV semantics
+	while ( abs( my $pid = waitpid( -1, POSIX ? WNOHANG : 0 ) ) > 1 ) {
+		_logmsg "reaped $pid" if DEBUG;
+	}
+	return;
+}
+
+
+our @pid;
+my $pid = $$;
+
+sub start_server {
+	my ( $self, $timeout ) = @_;
+	$timeout ||= 600;
+	croak 'Attempt to start ', ref($self), ' in a subprocess' unless $$ == $pid;
+	_logmsg('start server') if DEBUG;
+
+	foreach my $ip ( @{$self->{LocalAddr}} ) {
+		my $port = $self->{LocalPort};
+		push @pid, _spawn sub { $self->_TCP_server( $ip, $port, $timeout ) };
+		push @pid, _spawn sub { $self->_UDP_server( $ip, $port, $timeout ) };
+	}
+	return;
+}
+
+sub stop_server {
+	_logmsg('stop server') if DEBUG;
+	kill 'TERM', @pid;
+	return;
+}
+
+END {
+	local ( $!, $? );		## protect error and exit status
+	while ( abs( my $pid = waitpid( -1, 0 ) ) > 1 ) {
+		_logmsg "reaped $pid" if DEBUG;
+	}
+	_logmsg "terminated" if DEBUG;
 }
 
 
@@ -578,43 +584,65 @@ __END__
 
 =head2 new
 
-    $nameserver = Net::DNS::Nameserver->new(
-	LocalAddr	=> ['::1', '127.0.0.1'],
-	ZoneFile	=> "filename"
-	);
+	$nameserver = Net::DNS::Nameserver->new(
+			LocalAddr	=> ['::1', '127.0.0.1'],
+			LocalPort	=> 15353,
+			ZoneFile	=> "filename"
+			);
 
-    $nameserver = Net::DNS::Nameserver->new(
-	LocalAddr	=> '10.1.2.3',
-	LocalPort	=> 5353,
-	ReplyHandler	=> \&reply_handler,
-	Verbose		=> 1,
-	Truncate	=> 0
-    );
+	$nameserver = Net::DNS::Nameserver->new(
+			LocalAddr	=> '10.1.2.3',
+			LocalPort	=> 15353,
+			ReplyHandler	=> \&reply_handler,
+			Verbose		=> 1,
+			Truncate	=> 0
+			);
 
 Instantiates a Net::DNS::Nameserver object.
 An exception is raised if the object could not be created.
 
 Each instance is configured using the following optional arguments:
 
-    LocalAddr		IP address on which to listen	Defaults to loopback address
-    LocalPort		Port on which to listen		Defaults to 5353
-    ZoneFile		Name of file containing RRs
-			accessed using the internal
-			reply-handling subroutine
-    ReplyHandler	Reference to customised
-			reply-handling subroutine
-    NotifyHandler	Reference to reply-handling
-			subroutine for queries with
-			opcode NOTIFY (RFC1996)
-    UpdateHandler	Reference to reply-handling
-			subroutine for queries with
-			opcode UPDATE (RFC2136)
-    Verbose		Report internal activity	Defaults to 0 (off)
-    Truncate		Truncates UDP packets that
-			are too big for the reply	Defaults to 1 (on)
-    IdleTimeout		TCP clients are disconnected
-			if they are idle longer than
-			this duration			Defaults to 120 (secs)
+=over 4
+
+=item	LocalAddr
+
+IP address on which to listen.
+Defaults to the local loopback address.
+
+=item	LocalPort
+
+Port on which to listen.
+
+=item	ZoneFile
+
+Name of file containing RRs accessed using the internal reply-handling subroutine.
+
+=item	ReplyHandler
+
+Reference to customised reply-handling subroutine.
+
+=item	NotifyHandler
+
+Reference to reply-handling subroutine
+for queries with opcode NOTIFY (RFC1996).
+
+=item	UpdateHandler
+
+Reference to reply-handling subroutine
+for queries with opcode UPDATE (RFC2136).
+
+=item	Verbose
+
+Report internal activity.
+Defaults to 0 (off).
+
+=item	Truncate
+
+Truncates UDP packets that are too big for the reply.
+Defaults to 1 (on).
+
+=back
 
 The LocalAddr attribute may alternatively be specified as an array
 of IP addresses to listen to.
@@ -625,12 +653,33 @@ It must either return the response code and references to the answer,
 authority, and additional sections of the response, or undef to leave
 the query unanswered.  Common response codes are:
 
-    NOERROR	No error
-    FORMERR	Format error
-    SERVFAIL	Server failure
-    NXDOMAIN	Non-existent domain (name doesn't exist)
-    NOTIMP	Not implemented
-    REFUSED	Query refused
+=over 4
+
+=item NOERROR
+
+No error
+
+=item FORMERR
+
+Format error
+
+=item SERVFAIL
+
+Server failure
+
+=item NXDOMAIN
+
+Non-existent domain (name doesn't exist)
+
+=item NOTIMP
+
+Not implemented
+
+=item REFUSED
+
+Query refused
+
+=back
 
 For advanced usage it may also contain a headermask containing an
 hashref with the settings for the C<aa>, C<ra>, and C<ad>
@@ -643,108 +692,139 @@ EDNS options may be specified in a similar manner using the optionmask:
 See RFC1035 and IANA DNS parameters file for more information:
 
 
-The nameserver will listen for both UDP and TCP connections.
-On Unix-like systems, unprivileged users are denied access to ports below 1024.
+The nameserver will listen for both UDP and TCP connections.  On linux
+and other Unix-like systems, unprivileged users are denied access to
+ports below 1024.
 
-UDP reply truncation functionality was introduced in VERSION 830.
+UDP reply truncation functionality was introduced in Net::DNS 0.66.
 The size limit is determined by the EDNS0 size advertised in the query,
 otherwise 512 is used.
-If you want to do packet truncation yourself you should set C<Truncate>
-to 0 and truncate the reply packet in the code of the ReplyHandler.
-
-See L</EXAMPLE> for an example.
+If you want to do packet truncation yourself you should set Truncate=>0
+and truncate the reply packet in the code of the ReplyHandler.
 
 
-=head2 main_loop
+=head2 start_server
 
-    $ns->main_loop;
+	$ns->start_server( <TIMEOUT_IN_SECONDS> );
 
-Start accepting queries. Calling main_loop never returns.
+Starts a server process for each of the specified UDP and TCP sockets
+which continuously responds to user connections.
+
+The timeout parameter specifies the time the server is to remain active.
+If called with no parameter a default timeout of 10 minutes is applied.
 
 
-=head2 loop_once
+=head2 stop_server
 
-    $ns->loop_once( [TIMEOUT_IN_SECONDS] );
+	$ns->stop_server();
 
-Initialises the specified UDP and TCP sockets and starts the server
-which will respond to a single connection on each socket.
-
-The timeout parameter specifies the maximum time to wait for a request.
-If called without parameters a default timeout of 10 minutes is applied.
-If called with a zero parameter, the timeout function is disabled.
+Terminates all server processes in an orderly fashion.
 
 
-=head1 EXAMPLE
+=head1 EXAMPLES
 
-The following example will listen on port 5353 and respond to all queries
+=head2 Example 1: Test script with embedded nameserver
+
+The following example is a self-contained test script which queries DNS
+zonefile data served by an embedded Net::DNS::Nameserver instance.
+
+	use strict;
+	use warnings;
+	use Test::More;
+
+	plan skip_all => 'Net::DNS::Nameserver not available'
+			unless eval { require Net::DNS::Nameserver }
+			and Net::DNS::Nameserver->can('start_server');
+	plan tests => 2;
+
+	my $resolver = Net::DNS::Resolver->new(
+			nameserver => ['::1', '127.0.0.1'],
+			port	   => 15353
+			);
+	
+	my $ns = Net::DNS::Nameserver->new(
+			LocalAddr => [$resolver->nameserver],
+			LocalPort => $resolver->port,
+			Verbose	  => 0,
+			ZoneFile  => \*DATA
+			) or die "couldn't create nameserver object";
+
+	$ns->start_server(10);
+
+	my $reply = $resolver->send(qw(example.com SOA));
+	is( ref($reply), 'Net::DNS::Packet', 'received reply packet' );
+	my ($rr) = $reply->answer;
+	is( $rr->type, 'SOA', 'answer contains SOA record' );
+
+	$ns->stop_server();
+
+	exit;
+
+	__DATA__
+	$ORIGIN example.com.
+	@	IN SOA	mname rname 2023 2h 1h 2w 1h
+	www	IN A	93.184.216.34
+
+
+=head2 Example 2: Free-standing customised DNS nameserver
+
+The following example will listen on port 15353 and respond to all queries
 for A records with the IP address 10.1.2.3.  All other queries will be
 answered with NXDOMAIN.	 Authority and additional sections are left empty.
 The $peerhost variable catches the IP address of the peer host, so that
 additional filtering on a per-host basis may be applied.
 
-    #!/usr/bin/perl -T
+	use strict;
+	use warnings;
+	use Net::DNS::Nameserver;
 
-    use strict;
-    use warnings;
-    use Net::DNS::Nameserver;
+	sub reply_handler {
+		my ( $qname, $qclass, $qtype, $peerhost, $query, $conn ) = @_;
+		my ( $rcode, @ans, @auth, @add );
 
-    sub reply_handler {
-	my ( $qname, $qclass, $qtype, $peerhost, $query, $conn ) = @_;
-	my ( $rcode, @ans, @auth, @add );
+		print "Received query from $peerhost to " . $conn->{sockhost} . "\n";
+		$query->print;
 
-	print "Received query from $peerhost to " . $conn->{sockhost} . "\n";
-	$query->print;
+		if ( $qtype eq "A" && $qname eq "foo.example.com" ) {
+			my ( $ttl, $rdata ) = ( 3600, "10.1.2.3" );
+			my $rr = Net::DNS::RR->new("$qname $ttl $qclass $qtype $rdata");
+			push @ans, $rr;
+			$rcode = "NOERROR";
+		} elsif ( $qname eq "foo.example.com" ) {
+			$rcode = "NOERROR";
 
-	if ( $qtype eq "A" && $qname eq "foo.example.com" ) {
-		my ( $ttl, $rdata ) = ( 3600, "10.1.2.3" );
-		my $rr = Net::DNS::RR->new("$qname $ttl $qclass $qtype $rdata");
-		push @ans, $rr;
-		$rcode = "NOERROR";
-	} elsif ( $qname eq "foo.example.com" ) {
-		$rcode = "NOERROR";
+		} else {
+			$rcode = "NXDOMAIN";
+		}
 
-	} else {
-		$rcode = "NXDOMAIN";
+		# mark the answer as authoritative (by setting the 'aa' flag)
+		my $headermask = {aa => 1};
+
+		# specify EDNS options	{ option => value }
+		my $optionmask = {};
+
+		return ( $rcode, \@ans, \@auth, \@add, $headermask, $optionmask );
 	}
 
-	# mark the answer as authoritative (by setting the 'aa' flag)
-	my $headermask = {aa => 1};
+	my $ns = Net::DNS::Nameserver->new(
+			LocalPort    => 15353,
+			ReplyHandler => \&reply_handler,
+			Verbose	     => 1
+			) or die "couldn't create nameserver object";
 
-	# specify EDNS options	{ option => value }
-	my $optionmask = {};
+	$ns->start_server(60);
 
-	return ( $rcode, \@ans, \@auth, \@add, $headermask, $optionmask );
-    }
-
-
-    my $ns = Net::DNS::Nameserver->new(
-	LocalPort    => 5353,
-	ReplyHandler => \&reply_handler,
-	Verbose	     => 1
-	) || die "couldn't create nameserver object\n";
-
-
-    $ns->main_loop;
-
-    exit;
+	exit;	# leaving nameserver processes running for 60 seconds
 
 
 =head1 BUGS
 
-Limitations in perl make it impossible to guarantee that replies to
-UDP queries from Net::DNS::Nameserver are sent from the IP-address
-to which the query was directed.  This is a problem for machines with
-multiple IP-addresses and causes violation of RFC2181 section 4.
-Thus a UDP socket created listening to INADDR_ANY (all available
-IP-addresses) will reply not necessarily with the source address being
-the one to which the request was sent, but rather with the address that
-the operating system chooses. This is also often called "the closest
-address". This should really only be a problem on a server which has
-more than one IP-address (besides localhost - any experience with IPv6
-complications here, would be nice). If this is a problem for you, a
-work-around would be to not listen to INADDR_ANY but to specify each
-address that you want this module to listen on. A separate set of
-sockets will then be created for each IP-address.
+Limitations in perl make it impossible to guarantee that replies to UDP
+queries from Net::DNS::Nameserver are sent from the IP-address to which
+the query was directed, the source address being chosen by the operating
+system based upon its notion of "closest address". This limitation is
+mitigated to some extent by creating a separate socket and subprocess
+for each IP address.
 
 
 =head1 COPYRIGHT
@@ -757,7 +837,7 @@ Portions Copyright (c)2005 Robert Martin-Legene.
 
 Portions Copyright (c)2005-2009 O.M.Kolkman, RIPE NCC.
 
-Portions Copyright (c)2017,2023 R.W.Franks.
+Portions Copyright (c)2017-2024 R.W.Franks.
 
 All rights reserved.
 
@@ -787,10 +867,6 @@ L<perl> L<Net::DNS> L<Net::DNS::Resolver> L<Net::DNS::Packet>
 L<Net::DNS::Update> L<Net::DNS::Header> L<Net::DNS::Question>
 L<Net::DNS::RR>
 
-L<RFC1035|https://tools.ietf.org/html/rfc1035>
-
-L<IANA DNS parameters|http://www.iana.org/assignments/dns-parameters>
-
 =cut
 
 __END__
